/* scrnctl.c -- Screen Control module for KD.  Handles cursor movement
 *		scrolling, and other screen control functions.
 *
 *	Numerous bug fixes, mostly of VT100 emulation, cursor handling.
 *		Izumi Ohzawa, 3-2-90
 *	Back space fix
 *		Koichi Mori, 1-25-90
 *      Written to be compiled with Turbo C 2.0 (or later presumably) by
 *		Izumi Ohzawa, 1-8-90
 *		Group in Neurobiology/School of Optometry
 *		University of California, Berkeley
 *		Berkeley, CA 94720
 *		izumi@violet.berkeley.edu
 *
 *	This program/code is hereby released into the public domain.
 *	No copyright is claimed, and no restriction is imposed for any use,
 *	commercial or otherwise.  Use at your own risk.
 * ------------------------------------------------------------------------ */
/*
#define DEBUG 1
*/
#include <stdio.h>
#include <dos.h>
#include <stdlib.h>
#include <alloc.h>
#include <graphics.h>
#include <string.h>
#include <ctype.h>
#include "kjis.h"
#include "prototyp.h"

#define MAXSEQLEN	16	/* Maximum length of escape seq. not incl Esc */
#define	TABMASK		7	/* (Tab spacing -1) */
#define MAXESCSEQ	32	/* Array size for escape sequence */

#ifdef DEBUG
extern FILE *ffp;
#endif

/* Scroll area / Accessed in hercscrl.c too */
int Xul, Yul;	/* upper-left -- starts at (0,0) */
int Xbr, Ybr;	/* bottom-right */

/* Terminal State variables */
static int ESeqLen = 0;	/* current length of accumlating ESC sequence */
static int chcopy = 0;
static int Last1 = 0;
static int Last2 = 0;
static int ESC_ON =0;			/* Escape came */
static int K_IN = 0;			/* Kanji in flag */
static int LOW_BYTE = 0;		/* indicate next byte is low byte */
static unsigned int ASCorJIS = 0;	/* ASCII or JIS code buffer */
static char ESeqBuf[MAXESCSEQ];	/* escape sequence buffer */
static int XcurSave, YcurSave;		/* for cursor save/restore */

/* === Must put/erase cursor by Exclusive ORing in software */
void
Toggle_Cursor()
{
	lastxpos = Xcur<<3;
	lastypos = (Ycur+1)*PixelPerRow-1;
	putimage(lastxpos, lastypos, Cursor_Bitmap, XOR_PUT);
}


void
Clear_Box(int XUL, int YUL, int XBR, int YBR)
{
	Toggle_Cursor();
	setfillstyle(SOLID_FILL, 0);
	bar(XUL, YUL, XBR, YBR);
	Toggle_Cursor();
}


/* === Reset terminal state */
void
Reset_Term_State(void)
{
int i;
	Clear_All();			    /* fix for CGA.BGI's inability to clear */
	AscBackColor = 0;		    /* ASCII char background color */
	AscColor = DefAscColor;		    /* Reset current Ascii color (in kjis.h) */
	setcolor(AscColor);
	setfillstyle(SOLID_FILL, 0);
	ESeqLen = 0;
	for(i=0; i<MAXESCSEQ; i++)
		ESeqBuf[i] = 0;	
	chcopy = 0;
	Last1 = Last2 = 0;
	ESC_ON = K_IN = 0;
	LOW_BYTE = 0;
	ASCorJIS = 0;
	Set_Scroll_Area(0, 0, MaxColumn, MaxRow);
	Xcur = Ycur = 0;
	Toggle_Cursor();		/* fresh cursor at top left */
}


/* === Expand TAB character to appropriate number of spaces */
void
Expand_Tab(void)
{
	Toggle_Cursor();
	do
	{
		Xcur++;	
	} while(Xcur & TABMASK);
	if(Xcur>MaxColumn)
		Xcur = MaxColumn;
	Toggle_Cursor();
}

void
Set_Scroll_Area(int XXul, int YYul, int XXbr, int YYbr)
{
    if(XXul >= 0) Xul = XXul;
    else	  Xul = 0;
    if(YYul >= 0) Yul = YYul;
    else	  Yul = 0;
    if(XXbr <= MaxColumn) Xbr = XXbr;
    else		  Xbr = MaxColumn;
    if(YYbr <= MaxRow)    Ybr = YYbr;
    else		  Ybr = MaxRow;
}


/* === Scroll UP the specified area by NN line */
/* Arguments: Up/Down, # of lines, X,Y of upper-left, X,Y of bottom-right */
void
Scroll(int UpDown, int NN)
{
int i;
static union REGS inregs, outregs;
   if(ScrollMode == BIOS)
   {
	inregs.h.ah = UpDown;	/* UP=6, Down=7, Video Bios Scroll function */
        if(ScrollMult == 2)     /* CGA */
        {
	    inregs.h.al = NN<<1;                /* scroll 2*NN line */
	    inregs.h.dh = ((Ybr+1)<<1) -1;      /* for scroll, twice more lines */
        }
        else                    /* EGA, VGA */
        {
            inregs.h.al = NN;        
            inregs.h.dh = Ybr;
        }
	inregs.h.dl = Xbr;
	inregs.h.ch = Yul;
	inregs.h.cl = Xul;	/* CH:CL = row, column of upper left */
	inregs.h.bh = 0;
	int86(NEWVIDBIOS, &inregs, &outregs);
    }
    else	/* must be Hercules card */
    {		/* Currently, there's no definable scroll area for Hercules */
	if(UpDown == SCROLL_UP)
	    for(i=0; i<NN; i++)
		Herc_Scroll_Up();
	else	/* SCROLL_DOWN */
	    for(i=0; i<NN; i++)
		Herc_Scroll_Down();
    }
}


void
Back_Space(void)
{
	/* Back space should not delete any char.  Mod by KM */
	Toggle_Cursor();	/* take cursor off */
	if(Xcur>0)
	    Xcur--;
	Toggle_Cursor();
}

void
Do_CR(void)
{
	Toggle_Cursor();
	Xcur = 0;
	Toggle_Cursor();
}

void
Do_LF(void)
{
	Toggle_Cursor();
	if(Ycur < MaxRow)
		Ycur++;	/* empty line below */
	else
		Scroll(SCROLL_UP, 1);	/* already on bottom line */
	Toggle_Cursor();
}


void
New_Line(void)
{
	Toggle_Cursor();
        Xcur = 0;		/* CR */
	if(Ycur < MaxRow)	
		Ycur++;		/* empty line below */
	else
		Scroll(SCROLL_UP, 1);	/* already on bottom line */
	Toggle_Cursor();
}

void
Cursor_Up(int ScrollOk)
{
	Toggle_Cursor();
	if(Ycur > 0)
		Ycur--;		/* there is line above */
	else if(ScrollOk)
		Scroll(SCROLL_DOWN, 1);	/* already on bottom line */
	Toggle_Cursor();
}

void
Cursor_Down(int ScrollOk)
{
	Toggle_Cursor();
	if(Ycur < MaxRow)
		Ycur++;		/* there is line below */
	else if(ScrollOk)
		Scroll(SCROLL_UP, 1);	/* already on bottom line */
	Toggle_Cursor();
}

void
Cursor_Left(void)
{
	Toggle_Cursor();
	if(Xcur>0)
	    Xcur--;
	Toggle_Cursor();
}

void
Cursor_Right(void)
{
	Toggle_Cursor();
	if(Xcur<MaxColumn)
	    Xcur++;
	Toggle_Cursor();
}

void
Cursor_Moveto(int row, int column)
{
	if(column < 0) column = 0;
	if(column > MaxColumn) column = MaxColumn;
	if(row < 0) row = 0;
	if(row > MaxRow) row = MaxRow;
	Toggle_Cursor();
	Xcur = column;
	Ycur = row;
	Toggle_Cursor();
}


void
Save_Cursor(void)
{
	XcurSave = Xcur;
	YcurSave = Ycur;
}

void
Restore_Cursor(void)
{
	Toggle_Cursor();
	Xcur = XcurSave;
	Ycur = YcurSave;
	Toggle_Cursor();
}

void
Done_EscSeq(void)
{
int i;
	ESC_ON = 0;
	for(i=0; i<MAXESCSEQ; i++)
		ESeqBuf[i] = 0;	
	ESeqLen = 0;
}


/* === Atempt to emulate VT102/ANSI Escape sequence.  No guarantees */
void
Parse_Esc_Seq(int ch)
{
int irow, icolumn;
int xpos, ypos;

	/* accumulate Escape sequence in buffer ESeqBuf[], not incl. ESC itself */
	ESeqBuf[ESeqLen++] = ch;
	if(ESeqLen > MAXSEQLEN)
	{
	     Done_EscSeq();
	     return;
	}
	/* Test for sequences for each length */
	/* Current last char of esc sequence ESeqBuf[ESeqLen] is in 'ch' too */
	switch(ESeqLen)
	{
	    case 1:	/* 1-char Esc sequence */
		/* NOTE: First char of multi-char escape sequence must be HERE */
		if(ch == '$' || ch == '['  || ch == '#')
		    break;		/* Needed for 2 or more letter seq */
					/* So, do not aborb chars upto this */
		else if(ch == 'D')
		{
		    Toggle_Cursor();
		    Scroll(SCROLL_UP, 1);	/* Scroll text up */
		    Toggle_Cursor();
		}
		else if(ch == 'E')
		    New_Line();			/* as Newline */
		else if(ch == 'M')
		{
		    Toggle_Cursor();
		    Scroll(SCROLL_DOWN, 1);	/* Scroll text down */
		    Toggle_Cursor();
		}
		else if(ch == 'c')
		    Reset_Term_State();		/* RIS, reset terminal to ini. state */
		else if(ch == '7')
		    Save_Cursor();
		else if(ch == '8')
		    Restore_Cursor();
		else if(ch == '=')
		    {  /* Ignored: enter keypad application mode */ }
		else if(ch == '>')
		    {  /* Ignored: enter keypad numeric mode */ }
		
		Done_EscSeq();	/* reset ESC sequence interpretation */
		break;	/* Break from case 1: */

	    case 2:	/* 2-char Esc sequence */
		if(ESeqBuf[0] == '$')  /* <ESC>$.... */
		{
		    if(ch=='B' || ch=='@')	/* Kanji-In sequence */
		    {
			K_IN = 1;
			LOW_BYTE = 0;
			Toggle_Cursor();	/* No cursor in Kanji mode for speed */
			Done_EscSeq();
			break;
		    }
		}
		else if(ESeqBuf[0] == '[')
		{
		    /* NOTE: Second char of "[..." escape sequence must be HERE */
		    if((ch >='0' && ch <='9') || ch == ';' || ch == '?')
			break;			/* longer esc sequence */
		    else if(ch == 'A')		/* cursor up 1 line, no scroll */
			Cursor_Up(SCROLL_N);
		    else if(ch == 'B')		/* cursor down 1 line, no scroll */
		        Cursor_Down(SCROLL_N);
		    else if(ch == 'C')		/* cursor forward (right), stay on line */
			Cursor_Right();
		    else if(ch == 'D')		/* cursor backward (left), stay one line */
			Cursor_Left();
		    else if(ch == 'H')		/* home cursor, short for <esc>[0;0H */
		        Cursor_Moveto(0, 0);
		    else if(ch == 'K')		/* erase to EOL */
		    {
			xpos = Xcur<<3;
			ypos = Ycur*PixelPerRow;
		        Clear_Box(xpos, ypos, MaxX, ypos+15);
		    }
		    else if(ch == 'm')		/* turn off all attributes */
		    {
			AscColor = DefAscColor;	/* Reset current Ascii color (in kjis.h) */
			AscBackColor = 0;
			setcolor(AscColor);
		    }
		    else if(ch == 'r')		/* reset scrolling region to full screen */
			Set_Scroll_Area(0, 0, MaxColumn, MaxRow);
		    else if(ch == 's')
			Save_Cursor();
		    else if(ch == 'u')
			Restore_Cursor();

		    Done_EscSeq();	/* reset ESC sequence interpretation */
		    break;
		}
		else if(ESeqBuf[0] == '#')
		{
			/* none of this implemented */
			Done_EscSeq();	/* no other longer sequence */		
		}
		break;	/* Break from case 2: */

	    /* ------ 3 char sequences ---------------------------------- */
	    case 3:	/* Deal with 3 char escape sequences */
		xpos = Xcur<<3;
		ypos = Ycur*PixelPerRow;
		if(strcmp(ESeqBuf, "[;H") == 0)	/* Home cursor */
		    Cursor_Moveto(0, 0);
		else if(strcmp(ESeqBuf, "[2J") == 0)	/* Clear screen */
		    Reset_Term_State();
		else if(strcmp(ESeqBuf, "[1m") == 0)	/* Bold on */
		{
		  if(ColorOK)
		  {
		    AscColor = BoldColor;
		    AscBackColor = 0;
		    setcolor(AscColor);
		  }
		}
		else if(strcmp(ESeqBuf, "[7m") == 0)	/* Stand out */
		{
		  if(ColorOK)
		  {
		    AscColor = RevCharColor;	/* Do Reverse Video */
		    AscBackColor = RevBGColor;
		    setcolor(AscColor);
		  }
		}
		else if(strcmp(ESeqBuf, "[4m") == 0)	/* Underscore */
		{
		  if(ColorOK)
		  {
		    AscColor = EmphasisColor;
		    AscBackColor = 0;
		    setcolor(AscColor);
		  }
		}
		else if(strcmp(ESeqBuf, "[0K") == 0)	/* erase to EOL */
		    Clear_Box(xpos, ypos, MaxX, ypos+15);
	        else if(strcmp(ESeqBuf, "[1K") == 0)	/* erase from start to curs */
		    Clear_Box(0, ypos, xpos+7, ypos+15);
		else if(strcmp(ESeqBuf, "[2K") == 0)	/* erase whole line */
		    Clear_Box(0, ypos, MaxX, ypos+15);
		else
		    break;	/* if not the above, try more chars */
	
		Done_EscSeq();
		break;		/* end of 3-char escape sequence */

	    /* ------ Escape sequences of 4 or more chars after ESC ------ */
	    default:	/* Deal with other cases individually */
		if(ESeqLen > MAXSEQLEN)
		{
		    Done_EscSeq();
		    break;
		}
		else	/* Esq sequence length still OK */
		{
		   irow = icolumn = -1;
		   sscanf(ESeqBuf, "[%d;%dH", &irow, &icolumn);
		   if(irow != -1 && icolumn != -1)
		   {
		      irow--;			/* row=1..24 ->> 0..23 */
		      icolumn--;
		      if(ch == 'H' || ch == 'f')   /* <ESC>[%d;%dH -- cursor positioning */
		      {
		         Cursor_Moveto(irow, icolumn);
		         Done_EscSeq();
		         break;
		      }
		      else if(ch == 'r')   /* <ESC>[%d;%dr -- set scroll region */
		      {
			 if(SetScrollArea)
			 {
			     Set_Scroll_Area(0, irow, MaxColumn, icolumn);
		             		/* irow has top line */
			     		/* icolumn has bottom line */
			 }
		         Done_EscSeq();
		         break;
		      }
		   }  /* end of if(irow !=...) */

		   /* Try again with this special case format */
		   icolumn = -1;
		   sscanf(ESeqBuf, "[;%dH", &icolumn);
		   if(icolumn != -1)
		   {
		      icolumn--;
		      Cursor_Moveto(0, icolumn);
		      Done_EscSeq();
		      break;
		   }

		   /* Some mode set/reset sequences.. Just ignore them */
		   if(ESeqBuf[0] == '[' && ESeqBuf[1] == '?' && (ch == 'l' || ch == 'h'))
		   {
		      Done_EscSeq();
		      break;
		   }
		} /* end of last if(ESeqLen..){} else{.. */

		break;	/* Don't flush yet.  May partial, e.g., [15; may fail */
	}  /* end of switch(ESeqLen) */

}  /* end of Parse_Esc_Seq(int ch) */


/* === Process input stream 1 char at a time ========================= */
int
Process_One_Input(int ch)
{
	    if(ch == '\0')
		return(0);		/* ignore NULL */
	    Last2 = Last1;
	    Last1 = chcopy;
	    chcopy = ch;
	    if(ch == ESC_CHAR)
	    {
		ESC_ON = 1;
		return(0);
	    }

	    if(K_IN)   /* ########## If in Kanji Mode #################### */
	    {
		if(ch == '\n' || ch == '\f' || ch == CR_CHAR )
			return(0);	/* Don't accept LF, FF, CR during K_IN */
		/* check for Kanji-Out escape sequence */
		if(Last2==ESC_CHAR && Last1=='(' && (ch=='J' || ch=='B' || ch=='H'))
		{
		    K_IN = 0;
		    ESC_ON = 0;
		    if(Xcur <= MaxColumn)
			Toggle_Cursor();	/* restore cursor */
		    return(0);
		}
		if(LOW_BYTE)
		{
		    ASCorJIS += ch;
		    LOW_BYTE = 0;
		    Show_Char(ASCorJIS);
		}
		else   /* means high byte */
		{
		   ASCorJIS = ch<<8;
		   LOW_BYTE = 1;
		} /* end of high byte */
	    }	  /* end of if(K_IN) */
	    else  /* ########### In ASCII Mode ########################### */
	    {
		if(ch == '\n' || ch == '\f')
		{
		    Do_LF();
		    return(0);
		}
		if(ch == CR_CHAR)
		{
		    Do_CR();
		    return(0);
		}
		if(ESC_ON)
		{
		    Parse_Esc_Seq(ch);
		    return(0);
		}  /* end of if(ESC_ON) */
		else if(ch == BACKSPACE || ch == DEL_CHAR)
		    Back_Space();	/* will wipe out char to left */
		else if(ch == TAB_CHAR)
		    Expand_Tab();
		else	/* print normal ASCII char */
		    Show_Char((unsigned int)ch);

	    } /* end of if(K_IN){} else {} */
	    return(0);
}


